/******************************************************************************
*       SOFA, Simulation Open-Framework Architecture, development version     *
*                (c) 2006-2016 INRIA, USTL, UJF, CNRS, MGH                    *
*                                                                             *
* This library is free software; you can redistribute it and/or modify it     *
* under the terms of the GNU Lesser General Public License as published by    *
* the Free Software Foundation; either version 2.1 of the License, or (at     *
* your option) any later version.                                             *
*                                                                             *
* This library is distributed in the hope that it will be useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License *
* for more details.                                                           *
*                                                                             *
* You should have received a copy of the GNU Lesser General Public License    *
* along with this library; if not, write to the Free Software Foundation,     *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.          *
*******************************************************************************
*                               SOFA :: Modules                               *
* Authors: Igor Peterlik                                                      *
*                                                                             *
* Contact information: contact@sofa-framework.org                             *
******************************************************************************/
/*
Code contains An Algorithm for Automatically Fitting Digitized Curves
by Philip J. Schneider
from "Graphics Gems", Academic Press, 1990
*/

#include "BezierCubic.h"

#define MAXPOINTS 1000

namespace sofa {

namespace bezier4D {

Point4 BezierCubic::getPoint(double t)
{    
    Point4 	Q;	        // Point on curve at parameter t
    vector<Point4> 	Vtemp = controlPoints;		// Local copy of control points
    size_t degree = 3;

    // Triangle computation
    for (size_t i = 1; i <= degree; i++) {
        for (size_t j = 0; j <= degree-i; j++) {
            Vtemp[j][0] = (1.0 - t) * Vtemp[j][0] + t * Vtemp[j+1][0];
            Vtemp[j][1] = (1.0 - t) * Vtemp[j][1] + t * Vtemp[j+1][1];
            Vtemp[j][2] = (1.0 - t) * Vtemp[j][2] + t * Vtemp[j+1][2];
            Vtemp[j][3] = (1.0 - t) * Vtemp[j][3] + t * Vtemp[j+1][3];
        }
    }

    Q = Vtemp[0];
    return Q;
}

double BezierCubic::computeLength(size_t numPoints) {
    Point4 p1, p2;
    size_t subDiv = numPoints;
    length = 0.0;

    p2 = getPoint(0.0);
    for (size_t i = 1; i < subDiv; i++) {
        p1 = p2;
        p2 = getPoint(double(i)/double(subDiv));
        length += V4Dist3(p1, p2);
    }
    return length;
}

void BezierCubic::GenerateBezier(vector<CurvePoint4> &cnodes, int first, int last, vector<double> params, Vector4 tHat1,Vector4 tHat2)
{
    Vector4 	A[MAXPOINTS][2];	/* Precomputed rhs for eqn	*/
    double 	C[2][2];			/* Matrix C		*/
    double 	X[2];			/* Matrix X			*/
    double 	det_C0_C1,		/* Determinants of matrices	*/
            det_C0_X,
            det_X_C1;
    double 	alpha_l,		/* Alpha values, left and right	*/
            alpha_r;
    Vector4 	tmp;			/* Utility variable		*/

    bool optimizeCurve = false;


    if (verbose)
        std::cout << "F--L: " << first << " -- " << last << std::endl;

    size_t nPts = last - first + 1;

    curvePoints.clear();
    curvePoints.resize(nPts);
    curvePointsParam.clear();
    curvePointsParam.resize(nPts);
    controlPoints.clear();
    controlPoints.resize(4);
    tangents.clear();
    tangents.resize(2);

    for (size_t i = 0; i < nPts; i++) {
        curvePoints[i] = cnodes[first+i];
        curvePointsParam[i] = params[i];
    }

    tangents[0]=tHat1;
    tangents[1]=tHat2;

    /* Compute the A's	*/
    for (size_t i = 0; i < nPts; i++) {
        Vector4		v1, v2;
        v1 = tHat1;
        v2 = tHat2;
        V4ImposeLength(v1, B1(curvePointsParam[i]));
        V4ImposeLength(v2, B2(curvePointsParam[i]));
        A[i][0] = v1;
        A[i][1] = v2;
    }

    if (verbose)
        for (size_t i = 0; i < nPts; i++) {
            std::cout << curvePointsParam[i] << std::endl;
            std::cout << "A[" << i << "] = " << A[i][0] << " " << A[i][1] << std::endl;
        }

    /* Create the C and X matrices	*/
    if (optimizeCurve) {
        C[0][0] = 0.0;
        C[0][1] = 0.0;
        C[1][0] = 0.0;
        C[1][1] = 0.0;
        X[0]    = 0.0;
        X[1]    = 0.0;

        for (size_t i = 0; i < nPts; i++) {
            C[0][0] += V4Dot(A[i][0], A[i][0]);
            C[0][1] += V4Dot(A[i][0], A[i][1]);
            /*					C[1][0] += V4Dot(&A[i][0], &A[i][1]);*/
            C[1][0] = C[0][1];
            C[1][1] += V4Dot(A[i][1], A[i][1]);

            tmp = V4Sub(cnodes[first+i],
                        V4Add(
                            V4Scaled3(cnodes[first], B0(curvePointsParam[i])),
                            V4Add(
                                V4Scaled3(cnodes[first], B1(curvePointsParam[i])),
                                V4Add(
                                    V4Scaled3(cnodes[last], B2(curvePointsParam[i])),
                                    V4Scaled3(cnodes[last], B3(curvePointsParam[i]))
                                    )
                                )
                            )
                        );



            X[0] += V4Dot(A[i][0], tmp);
            X[1] += V4Dot(A[i][1], tmp);
        }



        /* Compute the determinants of C and X	*/
        det_C0_C1 = C[0][0] * C[1][1] - C[1][0] * C[0][1];
        det_C0_X  = C[0][0] * X[1]    - C[1][0] * X[0];
        det_X_C1  = X[0]    * C[1][1] - X[1]    * C[0][1];

        /* Finally, derive alpha values	*/
        alpha_l = (det_C0_C1 == 0) ? 0.0 : det_X_C1 / det_C0_C1;
        alpha_r = (det_C0_C1 == 0) ? 0.0 : det_C0_X / det_C0_C1;
    }


    /* If alpha negative, use the Wu/Barsky heuristic (see text) */
    /* (if alpha is 0, you get coincident control points that lead to
     * divide by zero in any subsequent NewtonRaphsonRootFind() call. */
    double segLength = V4Dist3(cnodes[last], cnodes[first]);
    double epsilon = 1.0e-6 * segLength;

    if (!optimizeCurve || (alpha_l < epsilon || alpha_r < epsilon || alpha_r > segLength || alpha_l > segLength)) {
        double dist = segLength / 3.0;
        controlPoints[0] = cnodes[first];
        controlPoints[3] = cnodes[last];
        controlPoints[1] = V4Add(controlPoints[0], V4ImposedLength(tHat1, dist));
        controlPoints[2] = V4Add(controlPoints[3], V4ImposedLength(tHat2, dist));

        if (verbose) {
            vector<Point4>& cp = controlPoints;
            std::cout << "CP1: " << std::endl;
            for (vector<Point4>::iterator ii = cp.begin(); ii != cp.end(); ii++)
                std::cout << ii->at(0) << " " << ii->at(1) << " " << ii->at(2) << std::endl;
        }

    } else {
        controlPoints[0] = cnodes[first];
        controlPoints[3] = cnodes[last];
        controlPoints[1] = V4Add(controlPoints[0], V4ImposedLength(tHat1, alpha_l));
        controlPoints[2] = V4Add(controlPoints[3], V4ImposedLength(tHat2, alpha_r));

        if (verbose) {
            vector<Point4>& cp = controlPoints;
            std::cout << "CP2: " << std::endl;
            std::cout << "Alpha l: " << alpha_l << std::endl;
            std::cout << "Alpha r: " << alpha_r << std::endl;
            for (vector<Point4>::iterator ii = cp.begin(); ii != cp.end(); ii++)
                std::cout << ii->at(0) << " " << ii->at(1) << " " << ii->at(2) << std::endl;
        }
    }
}

double BezierCubic::ComputeMaxError(vector<CurvePoint4>& cnodes, int first, int last, vector<double> params, int& splitPoint)
{
    int		i;
    double	dist;		//  Current error
    Point4	P;			//  Point on curve
    Vector4	v;			//  Vector from point to curve

    splitPoint = (last - first + 1)/2;
    double maxDist = 0.0;
    for (i = first + 1; i < last; i++) {
        P = BezierII(3, controlPoints, params[i-first]);
        v = V4Sub(P, cnodes[i]);
        dist = V4SquaredLength(v);
        if (dist >= maxDist) {
            maxDist = dist;
            splitPoint = i;
        }
    }
    return (maxDist);
}

void BezierCubic::ChordLengthParameterize(vector<CurvePoint4>& p, size_t first, size_t last, vector<double>& u)
{
    u.clear();
    u.resize(last-first+1);

    u[0] = 0.0;
    for (size_t i = first+1; i <= last; i++) {
        u[i-first] = u[i-first-1] + V4Dist3(p[i], p[i-1]);
    }

    for (size_t i = first + 1; i <= last; i++) {
        u[i-first] = u[i-first] / u[last-first];
    }
}

void BezierCubic::Reparameterize(vector<CurvePoint4>& p, int first,int last, vector<double>& params, vector<double>& newParams)
{
    int 	nPts = last-first+1;
    int 	i;

    //double	*uPrime;		/*  New parameter values	*/
    //uPrime = (double *)malloc(nPts * sizeof(double));

    newParams.clear();
    newParams.resize(nPts);

    for (i = first; i <= last; i++) {
        newParams[i-first] = NewtonRaphsonRootFind(p[i], params[i-first]);
    }
}


double BezierCubic::NewtonRaphsonRootFind(Point4 P, double params)
{
    double 		numerator, denominator;
    vector<Point4>   Q1(3), Q2(2);
    Point4		Q_u, Q1_u, Q2_u; /*u evaluated at Q, Q', & Q''	*/
    int 		i;

    /* Compute Q(u)	*/
    Q_u = BezierII(3, controlPoints, params);

    /* Generate control vertices for Q'	*/
    for (i = 0; i <= 2; i++) {
        Q1[i][0] = (controlPoints[i+1][0] - controlPoints[i][0]) * 3.0;
        Q1[i][1] = (controlPoints[i+1][1] - controlPoints[i][1]) * 3.0;
        Q1[i][2] = (controlPoints[i+1][2] - controlPoints[i][2]) * 3.0;
    }

    /* Generate control vertices for Q'' */
    for (i = 0; i <= 1; i++) {
        Q2[i][0] = (Q1[i+1][0] - Q1[i][0]) * 2.0;
        Q2[i][1] = (Q1[i+1][1] - Q1[i][1]) * 2.0;
        Q2[i][2] = (Q1[i+1][2] - Q1[i][2]) * 2.0;
    }

    /* Compute Q'(u) and Q''(u)	*/
    Q1_u = BezierII(2, Q1, params);
    Q2_u = BezierII(1, Q2, params);

    /* Compute f(u)/f'(u) */
    numerator = (Q_u[0] - P[0]) * (Q1_u[0]) + (Q_u[1] - P[1]) * (Q1_u[1]) + (Q_u[2] - P[2]) * (Q1_u[2]);
    denominator = (Q1_u[0]) * (Q1_u[0]) + (Q1_u[1]) * (Q1_u[1]) + (Q1_u[2]) * (Q1_u[2]) +
            (Q_u[0] - P[0]) * (Q2_u[0]) + (Q_u[1] - P[1]) * (Q2_u[1]) + (Q_u[2] - P[2]) * (Q2_u[2]);
    if (denominator == 0.0f) return params;

    /* u = u - f(u)/f'(u) */
    double newParam = params - (numerator/denominator);
    return (newParam);
}

/*bool BezierCubic::getRigidPoint(RigidCurvePoint4& _rigidPoint, double _t) {
    if (_t < 0.0d || _t > 1.0d) {
        std::cerr  << "BezierCurve::getRigidPoint: Cannot find a point for given parameter" << _t << std::endl;
        return(false);
    }

    CurvePoint4 p = getPoint(_t);
    Vector4 tang;
    if ( (1.0d - _t ) < tangentFactor)
        tang = V4Negated(tangents[1]);
    else {
        CurvePoint4 pt = getPoint(_t + tangentFactor);
        tang = V4Sub(pt,p);
    }
    _rigidPoint.set(p, tang);*/

/*std::cout << "From start of curve: " << _t*length << std::endl;
    CurvePoint4 p0 = getPoint(0.0d);
    CurvePoint4 pp = getPoint(1.0d);
    std::cout << "Linear distance: " << V4Dist3(p0, p) << std::endl;
    std::cout << "Length vs. linear distance: " << length  << " vs " << V4Dist3(p0, pp) << std::endl;*/

//return(true);
//}

bool BezierCubic::getRigidPoint(RigidCurvePoint4& _rigidPoint, double _t, const RigidCurvePoint4& _rigidPointPrevious) {
    if (_t < double(0.0) || _t > double(1.0) ) {
        std::cerr  << "BezierCurve::getRigidPoint: Cannot find a point for given parameter: " << _t << std::endl;
        return(false);
    }

    CurvePoint4 p = getPoint(_t);
    Vector4 tang;
    if ( (double(1.0) - _t ) < tangentFactor)
        tang = V4Negated(tangents[1]);
    else {
        CurvePoint4 pt = getPoint(_t + tangentFactor);
        tang = V4Sub(pt,p);
    }

    _rigidPoint.set(p, tang, _rigidPointPrevious);

    /*std::cout << "From start of curve: " << _t*length << std::endl;
    CurvePoint4 p0 = getPoint(0.0d);
    CurvePoint4 pp = getPoint(1.0d);
    std::cout << "Linear distance: " << V4Dist3(p0, p) << std::endl;
    std::cout << "Length vs. linear distance: " << length  << " vs " << V4Dist3(p0, pp) << std::endl;*/

    return(true);
}




}

}
