/******************************************************************************
*       SOFA, Simulation Open-Framework Architecture, development version     *
*                (c) 2006-2016 INRIA, USTL, UJF, CNRS, MGH                    *
*                                                                             *
* This library is free software; you can redistribute it and/or modify it     *
* under the terms of the GNU Lesser General Public License as published by    *
* the Free Software Foundation; either version 2.1 of the License, or (at     *
* your option) any later version.                                             *
*                                                                             *
* This library is distributed in the hope that it will be useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License *
* for more details.                                                           *
*                                                                             *
* You should have received a copy of the GNU Lesser General Public License    *
* along with this library; if not, write to the Free Software Foundation,     *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.          *
*******************************************************************************
*                               SOFA :: Modules                               *
* Authors: Igor Peterlik                                                      *
*                                                                             *
* Contact information: contact@sofa-framework.org                             *
******************************************************************************/

#include "BezierSegment.h"

namespace sofa {

namespace bezier4D {


BezierSegment::BezierSegment(vector<CurvePoint4> &cnodes, size_t _start, size_t _end, bool _knownStart, bool _knownEnd, string _vesselName, size_t _vesselID, size_t _segID, double _fittingError, bool _verbose) 
: vector<BezierCubic>() 
{
    size_t segStart, segEnd;

    segmentID = _segID;
    vesselName = _vesselName;
    vesselID = _vesselID;
    fittingError = _fittingError;
    verbose = _verbose;
    
    ID1 = cnodes[_start].topoID;
    ID2 = cnodes[_end].topoID;
    startCnode=_start;
    endCnode=_end;
    predCnode = cnodes[_start].pred;

//    std::cout << "HOLAStart segment cNode predecesor: " << cnodes[_start].pred << " ID: " << cnodes[cnodes[_start].pred].topoID <<std::endl;
//    std::cout << "Start: " << _start << " ID:" << cnodes[_start].topoID << " End: " << _end << " ID:" << cnodes[_end].topoID << std::endl;

    /*int inc = 0;
    if (!_knownStart && !_knownEnd) {
        //cout << "Segment open on both sides, assuming start as known" << endl;
        inc = 1;
        segStart = _start;
        segEnd = _end;
        isClosed = false;
    }

    if (_knownStart) {
        inc = 1;
        segStart = _start;
        segEnd = _end;

        if (_knownEnd) {
            //cout << "Segment closed on both sides." << endl;
            isClosed = true;
        } else {
            //cout << "Segment closed on start." << endl;
            isClosed = false;
        }
    }

    if (!_knownStart && _knownEnd)  {
        inc = -1;
        segStart = _end;
        segEnd = _start;
        isClosed = false;
    }*/

    isClosed = _knownStart && _knownEnd;

    if (verbose)
        std::cout << "Creating segment: " << _knownStart << " " << _knownEnd << " TopoIDs: " << ID1 << "--" << ID2 << " VesselID: " << vesselID << " SegmentID: " << segmentID << std::endl;

    if (_start >= _end) {
        std::cerr << "BezierSegment: error: first,last indices of segment: " << _start << " " << _end << std::endl;
        return;
    }

    segStart = _start;
    segEnd = _end;
    int inc = 1;

    /// compute the tangents on the right and left
    Vector4 tangLeft, tangRight;
    if (segStart == 0 || segStart+1 == cnodes.size() || (cnodes[segStart-inc].vID() != cnodes[segStart].vID()) ) {
        tangLeft = ComputeLeftTangent(cnodes[segStart], cnodes[segStart+inc]);
    } else {
        tangLeft = ComputeCenterTangent(cnodes[segStart-inc], cnodes[segStart], cnodes[segStart+inc]);
        V4Negate(tangLeft);
    }

    if ( segEnd == 0 || segEnd+1 == cnodes.size() || (cnodes[segEnd].vID() != cnodes[segEnd+inc].vID())) {
        tangRight = ComputeRightTangent(cnodes[segEnd-inc], cnodes[segEnd]);
    }
    else
        tangRight = ComputeCenterTangent(cnodes[segEnd-inc], cnodes[segEnd], cnodes[segEnd+inc]);

    clear();

    if (fittingError <= 0.0)
        fitSegmentRegular(cnodes, tangLeft, tangRight, segStart, segEnd);
    else
        fitSegmentAdaptive(cnodes, tangLeft, tangRight, segStart, segEnd);

    setCurveParams();

}

void BezierSegment::setCurveParams() {
    if (verbose)
        std::cout << "Number of curves in segment: " << size() << std::endl;

    size_t numCurveSubdiv = 1000;

    length = 0.0;
    for (size_t i = 0; i < size(); i++) {
        at(i).computeLength(numCurveSubdiv);        
        length+=at(i).length;
    }

    if (verbose)
        std::cout << "Length of segment: " << length << std::endl;

    double paramRight =  0.0;
    double paramLeft;
    double partLength = 0.0;
    for (size_t i = 0; i < size(); i++) {
        paramLeft = paramRight;
        partLength += at(i).length;
        paramRight = partLength/length;
        at(i).segParams.clear();
        at(i).segParams.resize(2);

        at(i).segParams[0] = paramLeft;
        at(i).segParams[1] = paramRight;

        if (verbose)
            std::cout << "Curve: " << i << " : " << paramLeft*length << " -- " << paramRight*length << std::endl;
    }
    
    if (fabs(1.0 - at(size()-1).segParams[1]) > 1e-4)
        std::cerr << "Right param of last curve not 1.0: " << at(size()-1).segParams[1] << std::endl;    
}

void BezierSegment::fitSegmentAdaptive(vector<CurvePoint4> &cnodes, Vector4& tang1, Vector4& tang2, size_t first, size_t last)  {
    double	maxError;	/*  Maximum fitting error	 */
    int		splitPoint;	/*  Point to split point set at	 */
    int		nPts;		/*  Number of points in subset  */
    double	iterationError; /*Error below which you try iterating  */
    int		maxIterations = 4; /*  Max times to try iterating  */
    Vector4	tHatCenter;   	/* Unit tangent vector at splitPoint */
    int		i;
    vector<double> params;

    if (verbose)
        std::cout << "Adaptively generating segment: " << first << " -- " << last << std::endl;

    iterationError = fittingError * fittingError;
    nPts = last - first + 1;

    /*  Use heuristic if region only has two points in it */
    if (nPts == 2) {
        if (verbose)
            std::cout << "Curve on 2 points " << std::endl;

        double dist = V4Dist3(cnodes[first], cnodes[last]) / 3.0;

        vector<CurvePoint4> p(2);
        p[0] = cnodes[first];
        p[1] = cnodes[last];

        vector<Point4> cp(4);
        cp[0] = p[0];
        cp[3] = p[1];

        vector<Point4> tangents(2);
        tangents[0] = V4ImposedLength(tang1, dist);
        tangents[1] = V4ImposedLength(tang2, dist);

        cp[1] = V4Add(cp[0], tangents[0]);
        cp[2] = V4Add(cp[3], tangents[1]);

        params.resize(2);
        params[0] = 0.0;
        params[1] = 1.0;

        BezierCubic bezCurve(p, params, cp, tangents, dist, verbose);
        push_back(bezCurve);

        return;
    }

    /*  Parameterize points, and attempt to fit curve */
    BezierCubic bezCurve(verbose);
    bezCurve.ChordLengthParameterize(cnodes, first, last, params);

    if (verbose) {
        std::cout << "Params = " ;
        for (vector<double>::iterator it = params.begin(); it != params.end(); it++)
            std::cout << *it << " ";
        std::cout << std::endl;
    }

    bezCurve.GenerateBezier(cnodes, first, last, params, tang1, tang2);

    /*  Find max deviation of points to fitted curve */
    maxError = bezCurve.ComputeMaxError(cnodes, first, last, params, splitPoint);

    if (verbose)
        std::cout << "Max error: " << maxError << std::endl;

    if (maxError < fittingError) {  /// accept the curve
        if (verbose)
            std::cout << "Accepting the curve for segment: " << first << " -- " << last << std::endl;
        push_back(bezCurve);
        return;
    }


    /*  If error not too large, try some reparameterization  */
    /*  and iteration */
    if (maxError < iterationError) {
        vector<double> newParams;
        for (i = 0; i < maxIterations; i++) {
            bezCurve.Reparameterize(cnodes, first, last, params, newParams);
            bezCurve.clear();
            bezCurve.GenerateBezier(cnodes, first, last, newParams, tang1, tang2);
            maxError = bezCurve.ComputeMaxError(cnodes, first, last, newParams, splitPoint);

            if (maxError < fittingError) {
                if (verbose)
                    std::cout << "Accepting the curve for segment after re-parametrization(" << maxIterations << "): " << first << " -- " << last << std::endl;
                push_back(bezCurve);
                return;
            }
            params=newParams;
        }
    }

    /* Fitting failed -- split at max error point and fit recursively */
    if (verbose)
        std::cout << "Recursive splitting" << std::endl;

    tHatCenter = ComputeCenterTangent(cnodes[first], cnodes[splitPoint], cnodes[last]);
    fitSegmentAdaptive(cnodes, tang1, tHatCenter, first, splitPoint);

    V4Negate(tHatCenter);
    fitSegmentAdaptive(cnodes, tHatCenter, tang2, splitPoint, last);
}


void BezierSegment::fitSegmentRegular(vector<CurvePoint4> &cnodes, Vector4 &tangLeft, Vector4 &tangRight, size_t segStart, size_t segEnd) {
    Vector4 t1, t2;
    for (size_type p = segStart; p != segEnd; p++) {
        if (p == segStart) {
            t1 = tangLeft;
        } else
            t1 = V4Negated(t2);

        if (p+1 == segEnd)
            t2 = tangRight;
        else
            t2 = ComputeCenterTangent(cnodes[p+1], cnodes[p+1], cnodes[p+2*1]);

        /// trivial chord parametrisation (from two points)
        vector<double> u(3);
        u[0] = 0.0;
        u[1] = u[2] = 1.0;

        vector<int> pi(2);
        pi[0] = p;
        pi[1] = p + 1;

        BezierCubic cubic(verbose);
        cubic.GenerateBezier(cnodes, p, p+1, u, t1, t2);
        push_back(cubic);
    }
}

/*bool BezierSegment::getPointFromDistance(RigidCurvePoint4& rigidPoint, double distanceFromStart) {
    CurvePoint4 p1, p2;
    Point3 tang;
    double tangentFactor = 0.001;
    double epsilon = 1e-6;
    double located = false;

    if ( fabs(distanceFromStart) < epsilon) {
        p1 = at(0).curvePoints[0];
        p2 = at(0).getPoint(tangentFactor);
        tang = getDiff4(p1, p2);
        located = true;
    } else {
        if (fabs(distanceFromStart - length) < epsilon) {
            p1 = at(size()-1).curvePoints[1];
            p2 = at(size()-1).getPoint(1-tangentFactor);
            tang = getDiff4(p2,p1);
            located = true;
        } else {
            double prevCumDist = 0.0, cumDist = 0.0;
            located = false;
            for (size_t cub = 0; cub < size(); cub++) {
                prevCumDist = cumDist;
                cumDist += at(cub).length;
                if (cumDist >= distanceFromStart) {
                    double ratio = (distanceFromStart - prevCumDist)/at(cub).length;
                    p1 = at(cub).getPoint(ratio);
                    if ( (1.0 - ratio) >= tangentFactor) {
                        p2 = at(cub).getPoint(ratio+tangentFactor);
                        tang = getDiff4(p1, p2);
                    } else {
                        p2 = at(cub).getPoint(ratio-tangentFactor);
                        tang = getDiff4(p2, p1);
                    }
                    located=true;
                    break;
                }
            }
        }
    }
    if (located)
        rigidPoint.set(p1 ,tang, distanceFromStart);
    return(located);
}*/

void BezierSegment::constructSampledPointsFromCurves() {
    sampledPoints.clear();
    //bool assigned = false;

    iterator it = begin();
    RigidCurvePoint4 rigidPoint;
    it->getRigidPoint(rigidPoint, double(0.0));
    rigidPoint.distanceFromStart=0.0;
    sampledPoints.push_back(rigidPoint);

    double totalLength=0.0;
    for (; it != end(); it++)  {
        RigidCurvePoint4 rigidPoint;
        it->getRigidPoint(rigidPoint, double(1.0), sampledPoints[sampledPoints.size()-1]);
        totalLength+=it->length;
        rigidPoint.distanceFromStart=totalLength;
        sampledPoints.push_back(rigidPoint);
    }

}

void BezierSegment::constructSampledPointsRegular(double distBetweenPoints) {
    double linkLength = length;
    size_t linkNum = 1;

    while (linkLength > distBetweenPoints) {
        linkNum++;
        linkLength = length/double(linkNum);
    }

    sampledPoints.clear();

    if (verbose)
        std::cout << "Segment length: " << length << " num links: " << linkNum << " link length: " << linkLength << std::endl;

    iterator it = begin();
    for (size_t i = 0; i <= linkNum; i++) {        
        double curvePos = double(i) * linkLength;                
        bool assigned = false;

        if (verbose)
            std::cout << "Adding point: D = " << curvePos << "(" << i << ")" << std::endl;

        size_t cv = 0;
        while (!assigned && it != end()) {
            double l = it->segParams[0]*length;
            double r = it->segParams[1]*length;

            if (fabs(curvePos - l) < ROUNDING_TOLERANCE)
                curvePos = l;

            if (fabs(curvePos - r) < ROUNDING_TOLERANCE)
                curvePos = r;

            if (verbose)
                std::cout << "   checking  interval: " << l << " -- " << r << std::endl;

            if (l <= curvePos && curvePos <= r) {
                double t =  (curvePos-l)/(r-l);
                RigidCurvePoint4 rigidPoint;

                if (verbose)
                    std::cout <<  "         |DS(" << cv << ")| = " << curvePos - l << "  t = " << t << std::endl;

                if (sampledPoints.size() > 0)
                    assigned = it->getRigidPoint(rigidPoint, t,  sampledPoints[sampledPoints.size()-1]);
                else /// to keep the orientation
                    assigned = it->getRigidPoint(rigidPoint, t);

                if (assigned) {
                    rigidPoint.distanceFromStart=curvePos;
                    sampledPoints.push_back(rigidPoint);
                    if (verbose)
                        std::cout << "      p = " << rigidPoint << std::endl;
                }
            }

            if (!assigned) {
                it++;
                cv++;
            }
        }

        if (!assigned) {
            std::cerr << "Point not assigned" << std::endl;
        }

    }
}


/*int BezierSegment::registerPoints(BezierSegment templSeg) {
    if (vesselID != templSeg.vesselID ||  segmentID != templSeg.segmentID) {
        cout << "Segments has different (vessel) ID, weird..." << vesselID << " vs. " << templSeg.vesselID << " " << segmentID << " vs. " << templSeg.vesselID << endl;
        return(-1);
    }

    if (isClosed != templSeg.isClosed) {
        cout << "Segments of different types (isClosed): " << isClosed << " vs. " << templSeg.isClosed << endl;
        return(-1);
    }

    double lengthRatio = 1.0;
    if (isClosed) {
        cout << "Registering closed segments: vesselID " << vesselID << " segmentID " << segmentID << endl;
        lengthRatio = templSeg.length/length;
        cout << "   Length ratio: " << lengthRatio << endl;
    }

    //for (size_t rp = 0; rp < sampledPoints.size(); rp++) {
    for (vector<RigidCurvePoint4>::iterator rpi = sampledPoints.begin(); rpi != sampledPoints.end(); rpi++) {
        double distance = lengthRatio * rpi->distanceFromStart;
        RigidCurvePoint4 rp;
        if (templSeg.getPointFromDistance(rp, distance))
            rpi->setMatchingPosition(rp);

    }

    return(1.0);
}*/

void BezierSegment::save(ofstream &file) {
    file << "#Segment " << vesselName << " " << vesselID << " " << segmentID << ( isClosed ? " C " : " O " ) << length << endl;

    file << "#ControlPoints " << size() << " " << size()*4 << endl;
    for (iterator cit = begin(); cit != end(); cit++)
        for (size_t p = 0; p < 4; p++) {
            for (size_t x = 0; x < 3; x++)
                file << " " << cit->controlPoints[p][x];
            file << endl;
        }

    file << "#MatchingPoints " << sampledPoints.size() << "  dist  x  y  z  w  qx qy qz qw R/N rx ry rz rw pred #succ [succ]" << endl;
    for (vector<RigidCurvePoint4>::iterator pit = sampledPoints.begin(); pit != sampledPoints.end(); pit++) {
        file << *pit;
    }
}

/*void BezierSegment::computeMapping(ChMesh *pMesh) {
    bool checkInsideTetra =  true;
    double circleRatioLowLimit = 0.0;
    double circleRatioUpLimit = 0.0;  /// 0.0  => no checking

    cout << "Segment mapping " << endl;

    for (size_t pi = 0; pi < sampledPoints.size(); pi++)  {
        RigidCurvePoint4& pt = sampledPoints[pi];

        if (!pt.hasMatchingPosition)
            continue;

        Point4 p(pt);
        //cout << p << endl;

        for (size_t el = 0; el < pMesh->nElems; el++) {

            //double df = 0.0;
            //for (size_t i = 0; i < 3; i++)
            //    df += SQR(pt[i] - pMesh->centers[el][i]);
            //df = sqrt(df);
            //if (df < 0.01) {
            //    cout << df << endl;
            //    cout << el << endl;
            //}

            vector<vector<double> >& vertices = pMesh->vecElementCoords[el];
            double **invVertMat = pMesh->invVertMatrices[el];

            bool accept = true;
            Point4 bc;
            //cout << "BC1= " << bc << endl;
            for (size_t i = 0; i < 3; i++)
                for (size_t j = 0; j < 3; j++)
                    bc[i] += invVertMat[i][j]*(pt[j] - vertices[3][j]);

            bc[3] = 1 - bc[0] - bc[1] - bc[2];

            double margin = 0.0;
            //double maxDiff = 0.0;
            for (size_t i = 0; i < 4; i++)
                accept = accept && (margin <= bc[i] && bc[i] <= 1.0-margin);

            //if (circleRatioUpLimit != 0.0) {
            //    double dst =  sqr(pMesh->innerCircles[el]);
            //    accept = accept && (dst*circleRatioLowLimit <= diff2 && diff2 < dst*circleRatioUpLimit);
            //}

            if (accept)  {
                if (pMesh->mappedElements.find(el) != pMesh->mappedElements.end())
                    cout << "Tetrahedron " << el << " already mapped " << endl;
                else {
                    pt.mappedTetra = el;
                    pt.baryCoord = bc;
                    pMesh->mappedElements.insert(el);



                    //cout << "Point [" << vesselID << "][" << segmentID << "][" << pi << "] mapped to " << el << " " <<  bc << endl;
                }
                break;
            }
        }
    }
}*/

}

}
