/******************************************************************************
*       SOFA, Simulation Open-Framework Architecture, development version     *
*                (c) 2006-2016 INRIA, USTL, UJF, CNRS, MGH                    *
*                                                                             *
* This library is free software; you can redistribute it and/or modify it     *
* under the terms of the GNU Lesser General Public License as published by    *
* the Free Software Foundation; either version 2.1 of the License, or (at     *
* your option) any later version.                                             *
*                                                                             *
* This library is distributed in the hope that it will be useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License *
* for more details.                                                           *
*                                                                             *
* You should have received a copy of the GNU Lesser General Public License    *
* along with this library; if not, write to the Free Software Foundation,     *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.          *
*******************************************************************************
*                               SOFA :: Modules                               *
* Authors: Igor Peterlik                                                      *
*                                                                             *
* Contact information: contact@sofa-framework.org                             *
******************************************************************************/
#ifndef BEZIER_SEGMENT_H
#define BEZIER_SEGMENT_H

#include "BezierHelper.h"
#include "BezierCubic.h"


namespace sofa {

namespace bezier4D {

#define ROUNDING_TOLERANCE double(1e-6)

class BezierSegment : public vector<BezierCubic>  {
public:    
    size_t vesselID;
    size_t segmentID;
    string vesselName;
    double length;
    bool isClosed;
    double fittingError;
    bool verbose;
    int startCnode, endCnode, predCnode;
    int ID1, ID2;
    

    vector<RigidCurvePoint4> sampledPoints;
    vector<int> sampledPtsSWCIdx;
    BezierSegment(vector<CurvePoint4> &cnodes, size_t start, size_t end, bool knownStart, bool knownEnd, string _vesselName, size_t _vesselID, size_t _segID, double _fittingError, bool _verbose = false);

    BezierSegment(vector<string> tokens, double scale = 1.0, bool _verbose = false) {
        verbose = _verbose;
        vesselName = tokens[1];
        vesselID = atoi(tokens[2].c_str());
        segmentID = atoi(tokens[3].c_str());
        if (tokens[4] == "C")
            isClosed = true;
        else
            isClosed = false;
        length = scale * atof(tokens[5].c_str());
    }

    BezierSegment(bool _verbose = false) : vector<BezierCubic>() {
        length = 0.0;
        ID1 = ID2 = -1;
        startCnode = endCnode = predCnode = -3;
        isClosed = false;
        verbose = _verbose;
    }

    void fitSegmentRegular(vector<CurvePoint4> &cnodes, Vector4& tangLeft, Vector4& tangRight, size_t segStart, size_t segEnd);
    void fitSegmentAdaptive(vector<CurvePoint4> &cnodes, Vector4 &tang1, Vector4 &tang2, size_t first, size_t last);
    
    void addAfterSegment(BezierSegment& _seg1) {
        if (_seg1.empty()) 
            _seg1 = *this;
        else {
            if (ID1 != _seg1.ID2) {
                std::cout << "Segments cannot be added: different IDs: seg1: " << _seg1.ID1 << "-" << _seg1.ID2 << "   seg2: " << ID1 << "-" << ID2 << std::endl; 
                return;
            } else {
                for (size_t ncur = 0; ncur < this->size(); ncur++)
                    _seg1.push_back(this->at(ncur));
                _seg1.length += length;
                _seg1.ID2 = ID2;
                _seg1.isClosed = _seg1.isClosed && isClosed;
                _seg1.fittingError = max(_seg1.fittingError, fittingError);
                _seg1.vesselID = -1;
                _seg1.segmentID = -1;
            }
        }
    }

    void setCurveParams();

    //int registerPoints(BezierSegment templSeg);
    //bool getPointFromDistance(RigidCurvePoint4& rigidPoint, double distance);

    void constructSampledPointsRegular(double distBetweenPoints);
    void constructSampledPointsFromCurves();
    void save(ofstream &file);

    //void computeMapping(ChMesh *pMesh);

    void printInfo() {
        cout << "Seg " << ID1 << " <--> " << ID2 << ( isClosed ? " Closed" : " Open" ) <<  " #C " << size() << " |D|: " << length << " V: " << vesselName << " #P " << sampledPoints.size() <<  endl;
    }
        

    void printsampledPoints() {
        //for (size_t p = 0; p < sampledPoints.size(); p++) {
        for (vector<RigidCurvePoint4>::iterator pit = sampledPoints.begin(); pit != sampledPoints.end(); pit++) {
            for (size_t i = 0; i < 4; i++)
                cout << " " << pit->at(i);
            cout << " ptID " << pit->topoID;
//            for (size_t i = 0; i < 4; i++)
//                cout << " " << pit->quat[i];
            if (pit->hasMatchingPosition)
                cout <<  " R" << pit->matchingPosition[0] << " " << pit->matchingPosition[1] << " " << pit->matchingPosition[2] << " " << pit->matchingPosition[3];
            else
                cout << " N - - -";
            //cout << " P[" << pit->mpIdPred.sID << "," << pit->mpIdPred.pID << "]";
            //for (size_t i = 0; i < pit->mpIdSucc.size(); i++)
            //    cout << " S[" << pit->mpIdSucc[i].sID << "," << pit->mpIdSucc[i].pID << "]";

            cout << endl;
        }
    }
};

}

}


#endif
