/******************************************************************************
*       SOFA, Simulation Open-Framework Architecture, development version     *
*                (c) 2006-2016 INRIA, USTL, UJF, CNRS, MGH                    *
*                                                                             *
* This library is free software; you can redistribute it and/or modify it     *
* under the terms of the GNU Lesser General Public License as published by    *
* the Free Software Foundation; either version 2.1 of the License, or (at     *
* your option) any later version.                                             *
*                                                                             *
* This library is distributed in the hope that it will be useful, but WITHOUT *
* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       *
* FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License *
* for more details.                                                           *
*                                                                             *
* You should have received a copy of the GNU Lesser General Public License    *
* along with this library; if not, write to the Free Software Foundation,     *
* Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA.          *
*******************************************************************************
*                               SOFA :: Modules                               *
* Authors: Igor Peterlik                                                      *
*                                                                             *
* Contact information: contact@sofa-framework.org                             *
******************************************************************************/

#ifndef SOFA_COMPONENT_BEZIER_SEGMENT_TREE_INL
#define SOFA_COMPONENT_BEZIER_SEGMENT_TREE_INL

#include "BezierSegmentTree.h"

#include <sofa/helper/system/glut.h>

#include <sofa/core/visual/VisualParams.h>
#include <sofa/core/behavior/MechanicalState.h>

#include <sofa/defaulttype/VecTypes.h>
#include <sofa/defaulttype/RigidTypes.h>
#include <sofa/defaulttype/Quat.h>
#include <sofa/simulation/Node.h>
#include <sofa/defaulttype/RigidTypes.h>
#include <sofa/defaulttype/VecTypes.h>
//#include <sofa/component/topology/Mesh2PointTopologicalMapping.h>

#include <sofa/core/visual/VisualParams.h>
#include <sofa/helper/gl/template.h>

#include <limits>

namespace sofa
{

namespace component
{

namespace misc
{

template<class DataTypes>
BezierSegmentTree<DataTypes>::BezierSegmentTree()
    : d_treeFile( initData(&d_treeFile, "treeFile", "Tree input file (obsolete format)") )
    , d_distanceBetweenPoints( initData(&d_distanceBetweenPoints, Real(1.0), "pointSamplingDistance", "Distance between sampled points: if 0: adaptive, if > 0: regular, if < 0: no sampling") )
    , d_collisionOffset( initData(&d_collisionOffset, Real(1.0), "collisionOffset", "distance between the first collision node and root of the tree (supposes linear tree, i.e. only one segment, WIP") )
    , d_positions( initData(&d_positions, "position", "Sampled positions") )
    , d_collisionPositions( initData(&d_collisionPositions, "collisionPosition", "3 DOF positions of nodes respecting the collision offset ") )
    , d_edges( initData(&d_edges, "edges", "Edges between sampled positions") )
    , d_collisionEdges( initData(&d_collisionEdges, "collisionEdges", "edges between collision positions") )
    , d_inputPointDataSegmentIDs( initData(&d_inputPointDataSegmentIDs, "inputPointDataSegmentIDs", "input segment sizes stored as point data (IDs)") )
    , d_inputTopology( initData(&d_inputTopology, "inputTopology", "Topology of the input tree") )
    , d_inputPositions( initData(&d_inputPositions, "inputPositions", "Positions of the input tree"))
    , d_pointRadii( initData(&d_pointRadii, "pointsRadii", "Radii of the tubes given in each inputPosition. If not specified, a constant value defaultRadius is used everywhere") )
    , d_fittingError( initData(&d_fittingError, double(0.0), "fittingError", "error for fitting curve, if <= 0.0, one curve per input point (not adaptative)"))
    , d_defaultRadius( initData(&d_defaultRadius, double(0.0), "defaultRadius", "default radius in case the vector of radii is not given"))
    , d_maxSegments( initData(&d_maxSegments, int(0), "maxSegments", "stop fitting curves after processing given number of segments"))
    , d_showCurveDiscretization( initData(&d_showCurveDiscretization, int(0), "showCurveDiscretization", "how many lines to draw the curve; if 0 no drawing"))
    , d_showInputPointSize( initData(&d_showInputPointSize, double(0.0), "showInputPointSize", "size of points used to fit the curves"))
    , d_showControlPointSize( initData(&d_showControlPointSize, double(0.0), "showControlPointSize", "size of Bezier control points"))
    , d_showTangentSize( initData(&d_showTangentSize, double(0.0), "showTangentSize", "size of tangents in boudary curve points"))
    , d_showSampledPointsSize( initData(&d_showSampledPointsSize, double(0.0), "showSampledPointsSize", "size of sampled points arrows"))
    , d_showCurve( initData(&d_showCurve, double(0.0), "showCurve", "show curves"))
    , d_xmlFilePrefix( initData(&d_xmlFilePrefix, "xmlFilePrefix", "prefix of generated XML file") )
    , d_swcFilePrefix( initData(&d_swcFilePrefix, "swcFilePrefix", "prefix of generated SWC graph file, only works correctly with subdivideBranchIntoSegments=false") )
    , d_verbose( initData(&d_verbose, false, "verbose", "verbose"))
    , d_cnodesIDFontSize( initData(&d_cnodesIDFontSize, "cpointsIDFontSize", "font size to visualize the curve-points"))
    , d_drawOffset( initData(&d_drawOffset, double(0.0), "drawOffset", "draw the tree with an offset in x-direction"))
    , d_treeCorrespondFile( initData(&d_treeCorrespondFile, "treeCorrespondFile", "file with segment correspondences") )
    , d_matchingPositions1( initData(&d_matchingPositions1, "matchingPositions1", "Sampled positions for matching segments1") )
    , d_matchingPositions2( initData(&d_matchingPositions2, "matchingPositions2", "Sampled positions for matching segments2") )
    , d_matchingEdges( initData(&d_matchingEdges, "matchingEdges", "Edges between sampled positions for matching (same for  target and source)") )
    , d_matchingEdgesIDs( initData(&d_matchingEdgesIDs, "matchingEdgesIDs", "Data associated to edges between sampled points, IDs of matched segments") )
    , d_matchingPointsIDs( initData(&d_matchingPointsIDs, "matchingPointsIDs", "Data associated to sampled points, IDs of matched segments") )
    , d_treeRoots( initData(&d_treeRoots, "treeRoot", "roots of the tree, i.e., the first N (5 for now) sampled point of the first segment") )
    , d_truncateOpenSegments( initData(&d_truncateOpenSegments, double(0.90), "truncateOpenSegments", "truncate open segments to avoid the curl at the end"))
    , d_bifurcationParentNodes( initData(&d_bifurcationParentNodes, "bifurcationParentNodes", "Indices of nodes that are \"parents\" in the bifurcations.") )
    , d_bifurcationChildNodes( initData(&d_bifurcationChildNodes, "bifurcationChildNodes", "Indices of nodes that are \"children\" in the bifurcations.") )
    , d_bifurcationPositions( initData(&d_bifurcationPositions, "bifurcationPositions", "Positions of bifurcations") )
    , d_subdivideInputSegments( initData(&d_subdivideInputSegments, true, "subdivideBranchIntoSegments", "chop the input branch to segmented w.r.t. bifurcations (increases the number of DOFs)"))
    , d_constructBifurcationPairs( initData(&d_constructBifurcationPairs, false, "exportBifurcationPairs", "export do data the parent-child index pairs; needed e.g. for constraints between branches"))
    , d_treeColor( initData(&d_treeColor, Vec3(-1.0,-1.0,-1.0), "treeColor", "color of the tree; if < 0, the tree segments will be randomly coloured"))
    , l_matchingBezierTreeLink(initLink("matchingBezierTree", "link to the another bezier tree for matching"))

{
    
}

template<class DataTypes>
BezierSegmentTree<DataTypes>::~BezierSegmentTree()
{
}

template<class DataTypes>
void BezierSegmentTree<DataTypes>::init() {
    size_t numPos = d_inputPositions.getValue().size();
    size_t numIDs = d_inputPointDataSegmentIDs.getValue().size();
    size_t numRad = d_pointRadii.getValue().size();

    if (numPos > 0) {
        /// positions on the input
        if (numIDs == 0) {
            /// we suppose that the points are organized in a single branch
            helper::WriteAccessor< Data< vector<int> > > pointDataIDs = d_inputPointDataSegmentIDs;
            pointDataIDs.resize(numPos,int(0));
            numIDs = numPos;
        }

        if (numRad != numPos) {
            double defRad = d_defaultRadius.getValue();
            if (numRad == 0) {
                std::cout << "Vector of radii not given. Using default value " << defRad << " everywhere." << std::endl;
                helper::WriteAccessor< Data< vector<double> > > pointsRadii = d_pointRadii;
                pointsRadii.resize(numPos);
                for (size_t ri = 0; ri < numPos; ri++)
                    pointsRadii[ri] = defRad;
            } else {
                std::cerr << "ERROR: numer of radii " << numRad << " differs from the number of points " << numPos << std::endl;
            }
        }

        if (numIDs == numPos) {
            std::cout << "Num points and IDs: " << numIDs << std::endl;
            createCNodesFromTopology();
        } else
            std::cerr << this->getName() << " ERROR: different size of positions and IDs" << std::endl;
    } else {
        if (d_treeFile.getValue() != "" ) {
            std::cout << "Reading data CSV files" << std::endl;
            readCNodesFromCSVFiles();
        } else {
            std::cerr << this->getName() << " ERROR: C-nodes cannot be created, no positions available" << std::endl;
            return;
        }
    }

    if (cnodes.size() > 0) {
        /// construct the c-nodes, i.e. add the topological information
        selectBranchingCNodes();
        /// fit curves to segments given by the c-nodes
        constructSegmentsFromCNodes();
        if (this->d_verbose.getValue())
            std::cout << this->getName() << " Number of bsegments: " << bezierSegments.size() << std::endl;
    } else
        std::cerr << this->getName() << " ERROR: No cnodes created" << std::endl;

    /// sample rigids along the Bezier segments
    if (d_distanceBetweenPoints.getValue() >= 0) {
        if (bezierSegments.size() > 0) {
            constructSampledPoints();
        } else
            std::cerr << this->getName() << " ERROR: No points sampled!" << std::endl;
    }

    if (d_constructBifurcationPairs.getValue())
        if (!createBifurcationIndices()) {
            std::cerr << this->getName() << " ERROR: bifurcation index pairs not exported!" << std::endl;
        }
    
    BezierSegmentTree<DataTypes>* matchingBezierTree = l_matchingBezierTreeLink.get();
    if (matchingBezierTree) {
        readTreeCorrespondences();
        matchSegmentsByRefitting();
        //matchBezierSegments();
    } else
        std::cout << this->getName() << " No matching will be done" << std::endl;


    /// generate colors for visualization:
    Vec3 tc = d_treeColor.getValue();
    if (tc[0] < 0 && tc[1] < 0 && tc[2] < 0) {
        for (size_t i = 0; i < 1000; i++) {
            rcol.push_back(double(rand()%1000) / double(1000.0));
            gcol.push_back(double(rand()%1000) / double(1000.0));
            bcol.push_back(double(rand()%1000) / double(1000.0));
        }
    } else {
        for (size_t i = 0; i < 1000; i++) {
            rcol.push_back(tc[0]);
            gcol.push_back(tc[1]);
            bcol.push_back(tc[2]);
        }
    }

//    if (bezierSegments.size() && cnodes.size() > 0) {
//        std::cout << "Save to XML file:" << d_xmlFilePrefix.getValue() << std::endl;
//        this->saveToXML(d_xmlFilePrefix.getValue());
//    }

    if (bezierSegments.size() && cnodes.size() > 0) {
//            std::cout << "Save to XML file:" << d_swcFilePrefix.getValue() << d_xmlFilePrefix.getValue() << std::endl;
            this->saveToSWC(d_swcFilePrefix.getValue());
    }
}

template<class DataTypes>
void BezierSegmentTree<DataTypes>::readTreeCorrespondences() {
    BezierSegmentTree<DataTypes>* matchingBezierTree = l_matchingBezierTreeLink.get();
    std::cout << "Bezier tree " << matchingBezierTree->getName() << " found for computing correspondences" << std::endl;
    string corrFile= d_treeCorrespondFile.getValue();
    ifstream fHandle;
    
    fHandle.open(corrFile.c_str());
    
    size_t in = 0;
    std::string line;
    char _sep = '-';
    
    segQualities.clear();
    while (fHandle.good() && !fHandle.eof()) {
        getline(fHandle, line);
        if (line.length() == 0)
            break;
        
        while (line[0] == '#') {
            std::cout << "First is # in " << line << std::endl;
            getline(fHandle, line);
            in++;
        }
        
        stringstream strstr(line);
        //std::cout << "Processing line " << line << std::endl;
        istream_iterator<std::string> it(strstr);
        istream_iterator<std::string> end;
        std::vector<std::string> tk(it, end);
        
        if (tk.size() < 2) {
            std::cerr << this->getName() << "ERROR: bad format, at least two words expected: X-Y-...-Z   A-B-...-C  ; line: " << in << std::endl;
            continue;
        } else {
            std::cout << this->getName() << "Line OK, the two words: " << tk[0] << " " << tk[1] << std::endl;
        }
        
        for (size_t wi = 0; wi < 2; wi++) {
            std::string w = tk[wi];
            for (size_t i = 0; i < w.length(); i++)   //naive substitution of separator
                if (w[i] == _sep)
                    w[i] = ' ';
            
            stringstream strstr2(w);
            istream_iterator<std::string> it2(strstr2);
            istream_iterator<std::string> end2;
            std::vector<std::string> tk2(it2, end2);
            
            vector<int> segs;
            
            for (size_t j = 0; j < tk2.size(); j++) {
                int ix = atoi(tk2[j].c_str());
                segs.push_back(ix);
            }
            
            matchingSegmentsIDs[wi].push_back(segs);
        }
        
        if (tk.size() > 2) {
            segQualities.push_back(tk[2]);
        } else
            segQualities.push_back("");
        
        double truncate = double(1.0);
        if (tk.size() > 3) {
            truncate = atof(tk[3].c_str());
        }
        segTruncate.push_back(truncate);
        
        in++;
    }
    fHandle.close();
    
    if (matchingSegmentsIDs[0].size() != matchingSegmentsIDs[1].size()) {
        std::cerr << "Number of correspond pairs differs: " << matchingSegmentsIDs[0].size() << " " << matchingSegmentsIDs[1].size() << std::endl;
        return;
    }
    
    /// print-out
    /*for (size_t i = 0; i < matchingSegmentsIDs[0].size(); i++) {
         for (size_t j = 0; j < matchingSegmentsIDs[0][i].size(); j++)
             std::cout << matchingSegmentsIDs[0][i][j] << " ";
         std::cout << "vs. ";
         for (size_t j = 0; j < matchingSegmentsIDs[1][i].size(); j++)
             std::cout << matchingSegmentsIDs[1][i][j] << " ";
         std::cout << std::endl;
     }*/
}

template<class DataTypes>
void BezierSegmentTree<DataTypes>::matchBezierSegments() {
    BezierSegmentTree<DataTypes>& matchingBezierTree = *l_matchingBezierTreeLink.get();
    matchSegments1.clear();
    matchSegments2.clear();

    helper::WriteAccessor< Data< VecCoord3D > > vecPos1 = d_matchingPositions1;
    helper::WriteAccessor< Data< VecCoord3D > > vecPos2 = d_matchingPositions2;
    helper::WriteAccessor< Data< vector<Edge> > > vecEdge = d_matchingEdges;
    helper::WriteAccessor< Data< vector<int> > > vecEdgeIDs = d_matchingEdgesIDs;
    helper::WriteAccessor< Data< vector<int> > > vecPointIDs = d_matchingPointsIDs;

    vecPos1.clear();
    vecPos2.clear();
    vecEdge.clear();
    vecEdgeIDs.clear();
    vecPointIDs.clear();

    std::cout << this->getName() << " # of segments to be processed: " << matchingSegmentsIDs[0].size() << std::endl;

    Real sampleDist = d_distanceBetweenPoints.getValue();

    std::cout << "Segments tree1: " << std::endl;
    for (size_t i = 0; i < bezierSegments.size(); i++)
        bezierSegments[i].printInfo();

    std::cout << "Segments tree1: " << std::endl;
    std::vector<BezierSegment> matchingBSegments = matchingBezierTree.getBezierSegments();
    for (size_t i = 0; i < matchingBSegments.size(); i++)
        matchingBSegments[i].printInfo();

    for (size_t spi = 0; spi < matchingSegmentsIDs[0].size(); spi++) {
        vector<int>& p0 = matchingSegmentsIDs[0][spi];
        vector<int>& p1 = matchingSegmentsIDs[1][spi];

        /// print out
        std::cout << this->getName() << " processing segment pair #" << spi << ":   ";
        for (size_t sp = 0; sp < p0.size(); sp++)
            std::cout << " " << p0[sp];
        std::cout << "  vs. ";
        for (size_t sp = 0; sp < p1.size(); sp++)
            std::cout << " " << p1[sp];
        std::cout << std::endl;
        if (segTruncate[spi]!=1.0)
            std::cout << "  truncating: " << segTruncate[spi] << std::endl;
    }

    /// TODO: now only the refitting variant is working and this version is to be completed
    /// perhaps it's the only feasible way in case there is N : M segment matchings
    /// in this case, doing the matching from the existing segments would be difficult and resampling is a better solution
}


template<class DataTypes>
void BezierSegmentTree<DataTypes>::matchSegmentsByRefitting() {
    BezierSegmentTree<DataTypes>* matchingBezierTree = l_matchingBezierTreeLink.get();
    matchSegments1.clear();
    matchSegments2.clear();
    
    helper::WriteAccessor< Data< VecCoord3D > > vecPos1 = d_matchingPositions1;
    helper::WriteAccessor< Data< VecCoord3D > > vecPos2 = d_matchingPositions2;
    helper::WriteAccessor< Data< vector<Edge> > > vecEdge = d_matchingEdges;
    helper::WriteAccessor< Data< vector<int> > > vecEdgeIDs = d_matchingEdgesIDs;
    helper::WriteAccessor< Data< vector<int> > > vecPointIDs = d_matchingPointsIDs;
    
    vecPos1.clear();
    vecPos2.clear();
    vecEdge.clear();
    vecEdgeIDs.clear();
    vecPointIDs.clear();
    
    std::cout << this->getName() << " # of segments to be processed: " << matchingSegmentsIDs[0].size() << std::endl;
    
    Real sampleDist = d_distanceBetweenPoints.getValue();
    
    for (size_t spi = 0; spi < matchingSegmentsIDs[0].size(); spi++) {
        vector<int>& p0 = matchingSegmentsIDs[0][spi];
        vector<int>& p1 = matchingSegmentsIDs[1][spi];
        
        /// print out
        std::cout << this->getName() << " processing segment pair #" << spi << ":   ";
        for (size_t sp = 0; sp < p0.size(); sp++)
            std::cout << " " << p0[sp];
        std::cout << "  vs. ";
        for (size_t sp = 0; sp < p1.size(); sp++)
            std::cout << " " << p1[sp];
        std::cout << std::endl;
        if (segTruncate[spi]!=1.0)
            std::cout << "  truncating: " << segTruncate[spi] << std::endl;

        CurveVec cnodes1, cnodes2;
        bool knownStart, knownEnd;
        
        /// matching tree segment fitting: redo fitting for matching segments
        for (size_t sp = 1; sp < p0.size(); sp++) {
            IPair key = make_pair(p0[sp-1], p0[sp]);
            std::map<IPair, IPair>::iterator valIt = matchingBezierTree->segmentIDsToIndices.find(key);
            
            if (valIt == matchingBezierTree->segmentIDsToIndices.end()) {
                std::cerr << " ERROR tree1: cannot find segment for IDs " << key.first << " " << key.second << std::endl;
                return;
            }
            
            //std::cout << this->getName() << " :subsegment " << key.first << " " << key.second << " has indices " << valIt->second.first << " " << valIt->second.second << std::endl;

            if (!matchingBezierTree->cnodes[valIt->second.first].isTopological() || !matchingBezierTree->cnodes[valIt->second.second].isTopological()) {
                std::cerr << this->getName() << " ERROR tree1: the first and the last cnode in the sub-segment must be topological! " << std::endl;
                return;
            }
            
            for (int cnode = valIt->second.first; cnode <= valIt->second.second; cnode++)
                cnodes1.push_back(matchingBezierTree->cnodes[cnode]);
        }
        size_t last1 = cnodes1.size()-1;
        knownStart = (cnodes1[0].type == 'S' || cnodes1[last1].type == 'C');
        knownEnd = (cnodes1[0].type == 'E' || cnodes1[last1].type == 'C');
        
        
        last1=size_t(last1*segTruncate[spi]);
        
        BezierSegment s1(cnodes1, 0, last1, knownStart, knownEnd, "OBSOLETE", 0, 0, matchingBezierTree->d_fittingError.getValue(), matchingBezierTree->d_verbose.getValue());
        
        /// matching tree segment fitting
        for (size_t sp = 1; sp < p1.size(); sp++) {
            IPair key = make_pair(p1[sp-1], p1[sp]);
            std::map<IPair, IPair>::iterator valIt = segmentIDsToIndices.find(key);
            
            if (valIt == segmentIDsToIndices.end()) {
                std::cerr << " ERROR tree2: cannot find segment for IDs " << key.first << " " << key.second << std::endl;
                return;
            }
            //std::cout << this->getName() << " :subsegment " << key.first << " " << key.second << " has indices " << valIt->second.first << " " << valIt->second.second << std::endl;
            if (!cnodes[valIt->second.first].isTopological() || !cnodes[valIt->second.second].isTopological()) {
                std::cerr << this->getName() << " ERROR tree2: the first and the last cnode in the sub-segment must be topological! " << std::endl;
                return;
            }
            
            for (int cnode = valIt->second.first; cnode <= valIt->second.second; cnode++)
                cnodes2.push_back(cnodes[cnode]);
        }
        size_t last2 = cnodes2.size()-1;
        last2=size_t(last2*segTruncate[spi]);
        knownStart = (cnodes2[0].type == 'S' || cnodes2[last2].type == 'C');
        knownEnd = (cnodes2[0].type == 'E' || cnodes2[last2].type == 'C');
        BezierSegment s2(cnodes2, 0, last2, knownStart, knownEnd, "OBSOLETE", 0, 0, d_fittingError.getValue(), d_verbose.getValue());
        
        s1.sampledPoints.clear();
        s2.sampledPoints.clear();
        
        std::cout << "     #  " << spi << " <" << s1.ID1 << " - " << s1.ID2 << " " << s1.length << " " << s1.isClosed << "> " <<
                     " <" << s2.ID1 << " - " << s2.ID2 << " " << s2.length << " " << s2.isClosed << "> " << segQualities[spi] << std::endl;
        
        if (s1.isClosed != s2.isClosed) {
            std::cerr << this->getName() << " ERROR: cannot match closed and open segments" << std::endl;
            return;
        }
        
        if (!s1.isClosed) {
            s1.constructSampledPointsRegular(sampleDist);
            s2.constructSampledPointsRegular(sampleDist);
        } else {
            double lengthRatio = s1.length/s2.length;
            std::cout << this->getName() << "   Matching CLOSED segments: l1=" << s1.length << " l2=" << s2.length << " l1/l2 = " << lengthRatio <<  std::endl;
            s1.constructSampledPointsRegular(sampleDist*lengthRatio);
            s2.constructSampledPointsRegular(sampleDist);
        }
        
        size_t maxSP = min(s1.sampledPoints.size(), s2.sampledPoints.size());
        s1.sampledPoints.resize(maxSP);
        s2.sampledPoints.resize(maxSP);
        
        
        for (size_t sp = 0; sp < s1.sampledPoints.size(); sp++) {
            Coord3D pos1, pos2;
            for (size_t d = 0; d < 3; d++)  {
                pos1[d] = s1.sampledPoints[sp][d];
                pos2[d] = s2.sampledPoints[sp][d];
            }
            vecPos1.push_back(pos1);
            vecPos2.push_back(pos2);
            vecPointIDs.push_back(spi);
            
            if (sp > 0) {
                Edge e(vecPos1.size()-2, vecPos1.size()-1);
                vecEdge.push_back(e);
                vecEdgeIDs.push_back(spi);
            }
        }
        
        //s1.printInfo();
        //s2.printInfo();
        
        matchSegments1.push_back(s1);
        matchSegments2.push_back(s2);
        
    }
}



template<class DataTypes>
void BezierSegmentTree<DataTypes>::bwdInit() {    
}

template<class DataTypes>
int BezierSegmentTree<DataTypes>::readCNodesFromCSVFiles() {
    fstream curvesFile;
    std::vector<string> curveNames, templateCurveNames, movingCurveNames;
    
    string inputFile = d_treeFile.getValue();
    curvesFile.open(inputFile.c_str());
    string path(".");
    
    while (curvesFile.good()) {
        string name;
        curvesFile >> name;
        if (name.size() > 4) {
            if (name.compare(name.size()-4, 4, ".csv") == 0) {
                sout << "Name = " << path+"/"+name << sendl;
                curveNames.push_back(path+"/"+name);
            } else
                path = name;
        }
    }
    curvesFile.close();
    
    size_t ncurves = curveNames.size();
    if ( ncurves == 0  ||  (ncurves%2) != 0) {
        cerr << "Bad curves file format" <<sendl;
        return(-1);
    }
    
    for (size_t i = 0; i <= (ncurves-1)/2; i++) {
        string argA(curveNames[2*i]);
        string argB(curveNames[2*i+1]);
        
        sout << "Template #" << i << ": " << argA << sendl;
        sout << "Moving   #" << i+1 << ": " << argB << sendl;
        
        templateCurveNames.push_back(argA);
        movingCurveNames.push_back(argB);
    }
    
    if (templateCurveNames.size() != movingCurveNames.size() ) {
        cerr << "Different number of template and moving vessels!" <<sendl;
        return(-2);
    } else
        sout << "Working with two trees, each having " << templateCurveNames.size() << " branches" <<sendl;
    
    std::vector<string>& _files = templateCurveNames;
    totalNumVessels = _files.size();
    
    ifstream fHandle;
    string line;
    char _sep = ',';
    
    /// read points from files, select  s/i/e
    for (size_t ves = 0; ves < _files.size(); ves++) {   // open a file
        fHandle.open(_files[ves].c_str(), ios::in);
        size_t in = 0;
        CurvePoint4 node;
        while (fHandle.good() && !fHandle.eof()) {
            getline(fHandle, line);
            if (line.length() == 0)
                break;
            
            if (line.at(0) == '\"')
                getline(fHandle, line);
            for (size_t i = 0; i < line.length(); i++)   //naive substitution of separator
                if (line.at(i) == _sep)
                    line.at(i)=' ';
            
            stringstream strstr(line);
            istream_iterator<std::string> it(strstr);
            istream_iterator<std::string> end;
            std::vector<std::string> tk(it, end);
            
            if (tk.size() != 3) {
                sout << "Problem in file " << _files[ves] << " line " << in <<sendl;
                return(-1);
            }
            for (size_t j = 0; j < 3; j++)
                node[j]=(atof(tk[j].c_str()));
            node[3] = 0.0;
            node.vID(ves);
            node.type = (in == 0) ? 's' : 'i';
            
            cnodes.push_back(node);
            in++;
        }
        cnodes[cnodes.size()-1].type = 'e';
        fHandle.close();
    }
    
    //for (size_t p = 0;  p < cnodes.size(); p++) {
    //    sout << "Point[" << p << "] " << cnodes[p][0] << " " << cnodes[p][1] << " " << cnodes[p][2];
    //    sout << " CV: " << cnodes[p].vesselID << " " << cnodes[p].type;
    //    sout <<sendl;
    //}
    return(0);
}

template<class DataTypes>
int BezierSegmentTree<DataTypes>::createCNodesFromTopology() {
    helper::ReadAccessor< Data< vector<Pos> > > inPos = d_inputPositions;
    helper::ReadAccessor< Data< vector<int> > > pointDataIDs = d_inputPointDataSegmentIDs;
    helper::ReadAccessor< Data< vector<double> > > pointsRadii = d_pointRadii;

    //std::cout << "Num of points, IDs, radii: " << inPos.size() << " " << pointDataIDs.size() << " " << pointsRadii.size() << std::endl;

    for (size_t nd = 0; nd < inPos.size(); nd++) {
        CurvePoint4 node(inPos[nd][0], inPos[nd][1], inPos[nd][2], pointsRadii[nd]);
        node.vID(pointDataIDs[nd]);
        
        if (nd == 0) {
            node.type = 's';
        } else {
            if (pointDataIDs[nd] != pointDataIDs[nd-1]) {
                node.type = 's';
                cnodes.back().type = 'e';
            } else
                node.type ='i';
        }
        cnodes.push_back(node);
    }
    cnodes.back().type = 'e';
    
    totalNumVessels = totalNumSegments = cnodes.back().vid+1; // inSegSizes.size();

    return(0);
}


template<class DataTypes>
int BezierSegmentTree<DataTypes>::selectBranchingCNodes() {
    /// find branching points from data:    
    for (size_t p = 1; p < cnodes.size(); p++) {
        for (size_t q = 0; q < p; q++) {
            if (V4Dist3(cnodes[p], cnodes[q]) < 1e-8) {
                //sout << "EQ " << p << " " << q <<sendl;
                cnodes[p].pred = q;
                
                switch (cnodes[p].type) {
                case 's': cnodes[p].type = 'S'; break;   /// known begin
                case 'e': cnodes[p].type = 'E'; break;   /// known end
                default: sout << "PROBLEM: start-branching of point " << p << " type: " << cnodes[p].type <<sendl; return(-1);
                }
                
                cnodes[q].succ.push_back(p);
                switch (cnodes[q].type) {
                case 'i': cnodes[q].type = 'C'; break;   /// branching point inside the vessel  (connective)
                case 's': cnodes[q].type = 'S'; break;   /// branching point as the starting point   (known begin)
                case 'e': cnodes[q].type = 'E'; break;   /// branching point as the ending point   (known end)
                case 'C':
                case 'S':
                case 'E': break;   /// already marked as branching
                default: sout << "PROBLEM: unknown type of branching point: " << q <<sendl; return(-1);
                }
                break;
            }
        }
    }
    
    
    /// create unique IDs for key points
    int ID=0;
    //for (CurveVec::iterator cit = cnodes.begin(); cit != cnodes.end(); cit++) {
    for (size_t ci = 0; ci < cnodes.size(); ci++) {
        CurvePoint4& cit = cnodes[ci];
        if (cit.isTopological() && cit.topoID == -1) {
            cit.topoID=ID++;
            for (vector<size_t>::iterator sit = cit.succ.begin(); sit != cit.succ.end(); sit++)
                cnodes[*sit].topoID = cit.topoID;
        }
    }

    //printCNodes();
    
    return(0);
}

template<class DataTypes>
BezierSegment* BezierSegmentTree<DataTypes>::findSegmentFromIDs(int ID1, int ID2) {
    BezierSegment* retVal = NULL;
    for (size_t i = 0; i < bezierSegments.size() && retVal == NULL; i++) {
        BezierSegment& bs = bezierSegments[i];
        if (bs.ID1 == ID1 && bs.ID2 == ID2)
            retVal = &bezierSegments[i];
        if (bs.ID1 == ID2 && bs.ID2 == ID1) {
            std::cout << "Inversed segment requested!" << std::endl;
            retVal = &bezierSegments[i];
        }
    }
    return retVal;
    
}

template<class DataTypes>
BezierSegment* BezierSegmentTree<DataTypes>::findSegmentFromPredCnode(int predCnode) {
    BezierSegment* retVal = NULL;
//    std::cout << "Find segment for node " << predCnode << std::endl;
    for (size_t i = 0; i < bezierSegments.size() && retVal == NULL; i++) {
        BezierSegment& bs = bezierSegments[i];
//        std::cout << "Look in segment " << i << " Start: " << bs.startCnode<< " End: " << bs.endCnode<< std::endl;
        if ( predCnode >= bs.startCnode &&  predCnode <= bs.endCnode ){
            retVal = &bezierSegments[i];
//            std::cout << "Segment found " << i << " ID1: "<< bs.ID1<< " ID2: "<< bs.ID2 << std::endl;
            break;
        }
//        if (bs.ID1 == ID2 && bs.ID2 == ID1) {
//            std::cout << "Inversed segment requested!" << std::endl;
//            retVal = &bezierSegments[i];
//        }
    }

    return retVal;

}


template<class DataTypes>
int BezierSegmentTree<DataTypes>::getPredIdxSWC(BezierSegment* predSeg, RigidCurvePoint4 predCnode) {
//    std::cout << "Get idx for predCnode " << predCnode[0] << " " << predCnode[1] << " " << predCnode[2] << " vID " << predCnode.vID() << " TopoID: " << predCnode.topoID << std::endl;
    std::vector<RigidCurvePoint4>& mpoints = predSeg->sampledPoints;
    std::vector<int>& sampledPtsIdx = predSeg->sampledPtsSWCIdx;
//    for(int sampJ=0;sampJ< sampledPtsIdx.size(); sampJ++){
//        std::cout << sampJ << " " << sampledPtsIdx[sampJ] << std::endl;
//        }
    double minDist=100000000;
    int closestPredIdx=-1;
    //int closestPi=0;
    for (size_t pi = 0; pi < mpoints.size(); pi++) {
        double dist = V4Dist3(mpoints[pi], predCnode);
        if (dist<minDist){
            minDist=dist;
            closestPredIdx=sampledPtsIdx[pi];
            //closestPi=pi;
        }
    }
//    std::cout << "Closest [" << closestPi << "] idxSWC[" << closestPredIdx << "] "<<  mpoints[closestPi][0] << " " << mpoints[closestPi][1] << " " << mpoints[closestPi][2] << " vID " << mpoints[closestPi].vID() << "TopoID: " << mpoints[closestPi].topoID << std::endl;
    return closestPredIdx;
}






template<class DataTypes>
void BezierSegmentTree<DataTypes>::printCNodes() {
    /// print points with the branching information
    for (size_t p = 0;  p < cnodes.size(); p++) {
        std::cout << "Point[" << p << "] " << cnodes[p][0] << " " << cnodes[p][1] << " " << cnodes[p][2];
        std::cout << " VID: " << cnodes[p].vID() << " T: " << cnodes[p].type << " I: " << cnodes[p].topoID;
        if (cnodes[p].succ.size() != 0)  {
            for (size_t q = 0; q < cnodes[p].succ.size(); q++)
                std::cout << " " << cnodes[p].succ[q];
        }
        if (cnodes[p].pred != -1)
            std::cout << " PRED: " << cnodes[p].pred;
        std::cout << std::endl;
    }
    
}


template<class DataTypes>
int BezierSegmentTree<DataTypes>::constructSegmentsFromCNodes() {
    /// create Bezier segments
    int start = -1;
    std::vector<size_t> segmentIDs(totalNumVessels,0);
    bool knownStart = false;
    
    bezierSegments.clear();
    
    segmentIDsToIndices.clear();
//    std::cout << "cNodes:"<<  std::endl;
    for (size_t p = 0; p < cnodes.size(); p++) {

//        std::cout << " TopoID: " << cnodes[p].topoID << " " << cnodes[p][0] << " " << cnodes[p][1] << " " << cnodes[p][2] << " pred: " << cnodes[p].pred << " vID " << cnodes[p].vID() << std::endl;
        size_t vesselID = cnodes[p].vID();
        string vesselName = "OBSOLETE";
        
        switch (cnodes[p].type) {
        case 's': start = p; knownStart = false; break;
        case 'S': start = p; knownStart = true; break;
        case 'i': break;
        case 'e': {
            //std::cout << this->getName() << ": create segment " << start << " -- " << p << " i.e. " << cnodes[start].topoID << " -- " << cnodes[p].topoID <<  std::endl;
            size_t truncatedEnd = start + size_t(d_truncateOpenSegments.getValue()*(p-start));
            //std::cout << "TRUNCATING CURVE FITTING by " << d_truncateOpenSegments.getValue() << " " << p << " -> " << truncatedEnd << std::endl;
            BezierSegment bsegment(cnodes, start, truncatedEnd, knownStart, false, vesselName, vesselID, segmentIDs[vesselID]++, d_fittingError.getValue(), d_verbose.getValue());
            bezierSegments.push_back(bsegment);
            IPair IDs(cnodes[start].topoID, cnodes[p].topoID);
            IPair IXs(start,p);
            segmentIDsToIndices.insert(std::make_pair(IDs,IXs));
            
            break;
        }
        case 'E': {
            //std::cout << this->getName() << ": create segment " << start << " -- " << p << " i.e. " << cnodes[start].topoID << " -- " << cnodes[p].topoID << std::endl;
            BezierSegment bsegment(cnodes, start, p, knownStart, true, vesselName, vesselID, segmentIDs[vesselID]++, d_fittingError.getValue(), d_verbose.getValue());
            bezierSegments.push_back(bsegment);
            IPair IDs(cnodes[start].topoID, cnodes[p].topoID);
            IPair IXs(start,p);
            segmentIDsToIndices.insert(std::make_pair(IDs,IXs));
            break;
        }
        case 'C': {
            //std::cout << this->getName() << ": create segment " << start << " -- " << p << " i.e. " << cnodes[start].topoID << " -- " << cnodes[p].topoID << std::endl;
            if (d_subdivideInputSegments.getValue()) {

                BezierSegment bsegment(cnodes, start, p, knownStart, true, vesselName, vesselID, segmentIDs[vesselID]++, d_fittingError.getValue(), d_verbose.getValue());
                bezierSegments.push_back(bsegment);
                IPair IDs(cnodes[start].topoID, cnodes[p].topoID);
                IPair IXs(start,p);
                segmentIDsToIndices.insert(std::make_pair(IDs,IXs));
                start = p;
                knownStart = true;
            }
            break;
        }
        default: {
            sout << "Problem, unknown type " << cnodes[p].type << " of a node " << p << sendl;
            return(-1);
        }
        }
        
        if (int(bezierSegments.size()) >= d_maxSegments.getValue()) {
            break;
        }
    }
//    std::cout << "HOLas "  << bezierSegments.size() << std::endl;
    return(0);
    
    /// update succ and pred according to (sID, pID)
}

template<class DataTypes>
bool BezierSegmentTree<DataTypes>::createBifurcationIndices() {
    helper::WriteAccessor< Data< vector<int> > > parInd = d_bifurcationParentNodes;
    helper::WriteAccessor< Data< vector<int> > > childInd = d_bifurcationChildNodes;
    helper::WriteAccessor< Data< VecCoord3D > > bfPos = d_bifurcationPositions;

    parInd.clear();
    childInd.clear();
    bfPos.clear();

    size_t bseg = 1;
    for (std::vector<BezierSegment>::iterator it=bezierSegments.begin()+1;  it != bezierSegments.end(); it++, bseg++) {
        if (it->sampledPoints.empty()) {
            std::cerr << "ERROR: no sampled points in Bezier segment " << bseg << std::endl;
            return false;
        }
        RigidCurvePoint4& fp = it->sampledPoints[0];
        Pos pch(fp[0], fp[1], fp[2]);
        //std::cout << "Bseg [" << bseg << "]: pos = " << pch << std::endl;
        int cIndex = fp.indexInMO;

        int pIndex = -1;
        Pos nbPos;
        for (std::vector<BezierSegment>::iterator itp=bezierSegments.begin(); pIndex == -1 && itp != it; itp++) {
            for (std::vector<RigidCurvePoint4>::iterator itpp = itp->sampledPoints.begin(); pIndex == -1 && itpp != itp->sampledPoints.end(); itpp++) {
                Pos pca(itpp->at(0), itpp->at(1), itpp->at(2));
                double dist = SQR(pch[0]-pca[0]) + SQR(pch[1]-pca[1]) + SQR(pch[2]-pca[2]);
                if (dist < 1e-10) {
                    pIndex = itpp->indexInMO;
                    nbPos = pca;
                }
            }

        }
        //std::cout << "Found combination:" << pIndex << " -> " << cIndex << std::endl;
        if (pIndex > -1 && cIndex > -1) {
            parInd.push_back(pIndex);
            childInd.push_back(cIndex);

            /// insert only unique positions of bifurcations
            bool insertPos = true;
            for (size_t bi = 0; bi < bfPos.size(); bi++) {
                Pos ebPos = bfPos[bi];
                double dist = sqrt(SQR(nbPos[0]-ebPos[0]) + SQR(nbPos[1]-ebPos[1]) + SQR(nbPos[2]-ebPos[2]));
                insertPos = insertPos && (dist > 1e-6);
            }
            if (insertPos)
                bfPos.push_back(nbPos);
        } else {
            std::cerr << "ERROR: bad parent -> child indices in segment " << bseg << ": " << pIndex << " -> " << cIndex << std::endl;
            return(false);
        }
    }
    return(true);
}


template<class DataTypes>
void BezierSegmentTree<DataTypes>::constructSampledPoints() {
    Real dist = d_distanceBetweenPoints.getValue();
    
    helper::WriteAccessor< Data< VecCoord > > posData = d_positions;
    //helper::WriteAccessor< Data< VecCoord3D > > posData3D = d_positions3D;
    helper::WriteAccessor< Data< vector<Edge> > > edgeData = d_edges;
    helper::WriteAccessor< Data< VecCoord > > treeRoots = d_treeRoots;

    size_t regPointInTree = 0;
    for (std::vector<BezierSegment>::iterator it=bezierSegments.begin();  it != bezierSegments.end(); it++) {
        if (dist > 0)
            it->constructSampledPointsRegular(double(dist));
        else
            it->constructSampledPointsFromCurves();

        size_t regPointInSegment = 0;
        
        for (std::vector<RigidCurvePoint4>::iterator itp = it->sampledPoints.begin(); itp != it->sampledPoints.end(); itp++) {
            Pos pos(itp->at(0), itp->at(1), itp->at(2));
            Quat quat(itp->quat[0], itp->quat[1], itp->quat[2], itp->quat[3]);
            
            Coord node(pos, quat);
            posData.push_back(node);
            //posData3D.push_back(pos);
            itp->indexInMO = posData.size()-1;

            if (regPointInTree <= 4)
                treeRoots.push_back(node);
            
            if (regPointInSegment > 0) {
                Edge e(regPointInTree-1, regPointInTree);
                edgeData.push_back(e);
            }
            
            regPointInSegment++;
            regPointInTree++;
        }
    }

    helper::WriteAccessor< Data< VecCoord3D > > collPos = d_collisionPositions;
    helper::WriteAccessor< Data< vector<Edge> > > collEdge = d_collisionEdges;

    size_t numPoints=posData.size();    

    /// temporary solution for filopodia with collisions: we suppose that the filopodium is given by single linear segment
    Coord3D rootPos = posData[0].getCenter();
    for (size_t i = 1; i < numPoints; i++) {
        Coord3D actPos = posData[i].getCenter();
        Coord3D diff = rootPos - actPos;

        if (  diff.norm() > d_collisionOffset.getValue())
            collPos.push_back(actPos);
    }

    for (size_t i = 1; i < collPos.size(); i++) {
        Edge e(i-1, i);
        collEdge.push_back(e);
    }


    /*posDataWL.resize(numPoints-1);
    edgeDataWL.resize(numEdges-1);
    posDataL.resize(1);

    for (size_t i = 0; i < numEdges-1; i++) {
        edgeDataWL[i] = edgeData[i];
        posDataWL[i] = posData[i];
    }

    posDataWL[numPoints-2] = posData[numPoints-2];
    posDataL[0] = posData[numPoints-1];*/

}

template<class DataTypes>
void BezierSegmentTree<DataTypes>::saveToXML(string filePrefix) {    
    helper::ReadAccessor< Data< VecCoord > > vecPos = d_positions;
    
    fstream fPoints, fTopo; //, fRegPoints, fSource, fTarget;
    
    //string fileName = filePrefix+"SourceMO.xml";
    //fSource.open(fileName.c_str(), ios::out);
    
    //fileName = filePrefix+"TargetMO.xml";
    //fTarget.open(fileName.c_str(), ios::out);
    
    string fileName = filePrefix+"RigidMO.xml";
    fPoints.open(fileName.c_str(), ios::out);
    
    fileName=filePrefix+"TOPO.xml";
    fTopo.open(fileName.c_str(), ios::out);
    
    //fileName=regPrefix+"TargetRigidMO.xml";
    //fRegPoints.open(fileName.c_str(), ios::out);
    
    fPoints << "<MechanicalObject template=\"Rigid\" scale3d=\"scale3d\" name=\"BeamMO\" position=\"";
    //fRegPoints << "<MechanicalObject template=\"Rigid\" scale3d=\"scale3d\" name=\"RegMO\" position=\"" << endl;
    
    //fSource<< "<MechanicalObject template=\"Vec3d\" scale3d=\"scale3d\" name=\"SourceMO\" position=\"" << endl;
    //fTarget << "<MechanicalObject template=\"Vec3d\" scale3d=\"scale3d\" name=\"TargetMO\" position=\"" << endl;
    
    fTopo << "<Mesh name=\"BeamTopology\" lines=\""<< endl;
    sout << "SIZE: " << bezierSegments.size() << sendl;
    
    for (size_t pi = 0; pi < vecPos.size(); pi++)
        fPoints << std::endl << "\t\t\t" << vecPos[pi];
    
    size_t regPointInTree = 0;
    for (size_t si = 0; si < bezierSegments.size(); si++) {
        std::vector<RigidCurvePoint4>& mpoints = bezierSegments[si].sampledPoints;
        fTopo << "\t\t\t\t";
        size_t regPointInSegment = 0;
        //sout << "Segment size: " << mpoints.size() << sendl;
        for (size_t pi = 0; pi < mpoints.size(); pi++) {
            //if (mpoints[pi].hasMatchingPosition) {
            //if (mpoints[pi].mappedTetra != -1) {
            //fPoints << "\t\t\t\t" << mpoints[pi][0] << " " << mpoints[pi][1] << " " << mpoints[pi][2] << " ";
            //fPoints << mpoints[pi].quat[0] << " " << mpoints[pi].quat[1] << " " << mpoints[pi].quat[2] << " " << mpoints[pi].quat[3] << endl;
            
            //fRegPoints << "\t\t\t\t" << mpoints[pi].matchingPosition[0] << " " << mpoints[pi].matchingPosition[1] << " " << mpoints[pi].matchingPosition[2] << " ";
            //fRegPoints << " 0 0 0 1" << endl;
            
            //fSource<< "\t\t\t\t" << mpoints[pi][0] << " " << mpoints[pi][1] << " " << mpoints[pi][2] << " " << endl;
            //fTarget<< "\t\t\t\t" << mpoints[pi].matchingPosition[0] << " " << mpoints[pi].matchingPosition[1] << " " << mpoints[pi].matchingPosition[2] << " " << endl;
            
            
            if (regPointInSegment > 0) {
                fTopo  << regPointInTree-1 << " " << regPointInTree << " ";
            }
            regPointInSegment++;
            regPointInTree++;
            //}
        }
        fTopo << endl;
    }
    
    //fTarget << "\t\t\t\t\"/>" << endl;
    //fSource << "\t\t\t\t\"/>" << endl;
    fPoints << "\t\t\t\t\"/>" << endl;
    //fRegPoints << "\t\t\t\t\"/>" << endl;
    fTopo << endl << "\t\t\t\t\"/>" << endl;
    
    //fTarget.close();
    //fSource.close();
    fPoints.close();
    //fRegPoints.close();
    fTopo.close();
}




template<class DataTypes>
void BezierSegmentTree<DataTypes>::saveToSWC(string filePrefix) {

    if(filePrefix.empty()){
        return;
    }

    std::cout << "Save bezier interpolated to SWC graph file:" << filePrefix << std::endl;

//    helper::ReadAccessor< Data< VecCoord > > vecPos = d_positions;
    fstream fTopo;
    string fileName=filePrefix+"Interpolated.swc";
    fTopo.open(fileName.c_str(), ios::out);
    sout << "Save to SWC file:" << fileName << std::endl;
    //    for (size_t pi = 0; pi < vecPos.size(); pi++)
//        fTopo << std::endl << "\t\t\t" << vecPos[pi];

    size_t regPointInTree = 0;size_t swcIdx = 1; int cellType=1;
    for (size_t si = 0; si < bezierSegments.size(); si++) {
//        sout << "write SWC segment: " << si << sendl;
        std::vector<RigidCurvePoint4>& mpoints = bezierSegments[si].sampledPoints;
        std::vector<int>& sampledPtsIdx = bezierSegments[si].sampledPtsSWCIdx;
//        bezierSegments[si].printsampledPoints();

//        fTopo << "HOLAS\n";
        size_t regPointInSegment = 0;

        for (size_t pi = 0; pi < mpoints.size(); pi++) {
            sampledPtsIdx.push_back(swcIdx);
            if(si==0 && regPointInSegment==0){
                fTopo  << swcIdx<<" " << cellType << " " << mpoints[pi][0] << " " << mpoints[pi][1] << " " << mpoints[pi][2] << " " << mpoints[pi][3]<< " -1"  << endl;
            }
            else if (regPointInSegment > 0) {
                fTopo  << swcIdx<< " " << cellType << " "<< mpoints[pi][0] << " " << mpoints[pi][1] << " " << mpoints[pi][2] << " " << mpoints[pi][3] << " " << swcIdx-1 << endl;
                 //sampledPtsIdx.push_back(swcIdx);
            }
            else{
                if (bezierSegments[si].predCnode<=0){
                    break;
                }
                BezierSegment* predSeg=findSegmentFromPredCnode(bezierSegments[si].predCnode);

                if(predSeg==NULL){
                     break;
                }
                int predSegSwcIdx=getPredIdxSWC(predSeg, mpoints[pi]);
                fTopo  << swcIdx<< " "<< cellType << " " << mpoints[pi][0] << " " << mpoints[pi][1] << " " << mpoints[pi][2] << " " << mpoints[pi][3] << " " << predSegSwcIdx << endl;
                //sampledPtsIdx.push_back(-2);
            }

            swcIdx++;
            regPointInSegment++;
            regPointInTree++;
            if (swcIdx>3){
                cellType=2;
            }
            //}
        }
//        fTopo << endl;
//        radius+=2.0;
    }

//    fTopo << endl;
    fTopo.close();
}






template<class DataTypes>
void BezierSegmentTree<DataTypes>::draw(const core::visual::VisualParams* vparams) {
    //if (!vparams->displayFlags().getShowVisualModels()) return;
    
    glDisable(GL_LIGHTING);
    
    int cPoints = d_showCurveDiscretization.getValue();
    
    double drawOffset = d_drawOffset.getValue();
    
    /// write out the cnode unique IDs
    if (this->d_cnodesIDFontSize.getValue() > 0.0 && vparams->displayFlags().getShowBoundingCollisionModels()) {
        glPointSize(1.0);
        float scale = float(this->d_cnodesIDFontSize.getValue()); // (float)( ( vparams->sceneBBox().maxBBox() - vparams->sceneBBox().minBBox() ).norm() * this->branchIDFontSize.getValue());
        Mat<4,4, GLfloat> modelviewM;
        //for (size_t pt = 0; pt < cnodes.size(); pt++) {
        for (CurveVec::iterator cit = cnodes.begin(); cit != cnodes.end(); cit++) {
            if (!cit->isTopological())
                continue;
            
            Vec3d tc;
            
            for (size_t ix = 0; ix < 3; ix++)
                tc[ix] = cit->at(ix);
            tc[0] += drawOffset;
            
            std::ostringstream oss;
            oss << cit->topoID; // << cit->type;
            std::string tmp = oss.str();
            const char* s = tmp.c_str();
            
            glColor3f(0.0f,1.0f,0.0f);
            
            glPushMatrix();
            
            glTranslatef(float(tc[0]), float(tc[1]), float(tc[2]));
            glScalef(scale,scale,scale);
            
            // Makes text always face the viewer by removing the scene rotation
            // get the current modelview matrix
            glGetFloatv(GL_MODELVIEW_MATRIX , modelviewM.ptr() );
            modelviewM.transpose();
            
            Vec3d temp = tc;
            temp = modelviewM.transform(temp);
            
            //glLoadMatrixf(modelview);
            glLoadIdentity();
            
            glTranslatef(float(temp[0]), float(temp[1]), float(temp[2]));
            glScalef(scale,scale,scale);
            
            while(*s)
            {
                glutStrokeCharacter(GLUT_STROKE_ROMAN, *s);
                s++;
            }
            
            glPopMatrix();
        }
    }

    /// show segment matching
    if (vparams->displayFlags().getShowInteractionForceFields()) {
        for (size_t seg = 0; seg < matchSegments1.size(); seg++) {
            //glColor4f(rcol[seg],gcol[seg],bcol[seg],1);
            glColor4f(0.0,1.0,1.0,1.0);
            for (size_t cv = 0; cv < matchSegments1[seg].size(); cv++) {
                Point4 pPrev = matchSegments1[seg][cv].getPoint(0.0);
                Point4 pAct;

                for (int p = 1; p <= cPoints; p++) {

                    double t = double(p)/double(cPoints);
                    pAct = matchSegments1[seg][cv].getPoint(t);

                    Pos pt1(pPrev[0], pPrev[1], pPrev[2]);
                    Pos pt2(pAct[0], pAct[1], pAct[2]);

                    glLineWidth(d_showCurve.getValue());
                    glBegin(GL_LINES);
                    helper::gl::glVertexT(pt1);
                    helper::gl::glVertexT(pt2);
                    glEnd();

                    pPrev = pAct;
                }
            }

            for (size_t cv = 0; cv < matchSegments2[seg].size(); cv++) {
                Point4 pPrev = matchSegments2[seg][cv].getPoint(0.0);
                Point4 pAct;

                for (int p = 1; p <= cPoints; p++) {

                    double t = double(p)/double(cPoints);
                    pAct = matchSegments2[seg][cv].getPoint(t);

                    Pos pt1(pPrev[0]+drawOffset, pPrev[1], pPrev[2]);
                    Pos pt2(pAct[0]+drawOffset, pAct[1], pAct[2]);

                    glLineWidth(d_showCurve.getValue());
                    glBegin(GL_LINES);
                    helper::gl::glVertexT(pt1);
                    helper::gl::glVertexT(pt2);
                    glEnd();

                    pPrev = pAct;
                }
            }

            std::vector<RigidCurvePoint4>& sp1 = matchSegments1[seg].sampledPoints;
            std::vector<RigidCurvePoint4>& sp2 = matchSegments2[seg].sampledPoints;

            if (sp1.size() != sp2.size()) {
                //std::cerr << "Different number of sampled points for matching segments!" << std::endl;
            } else {
                for (size_t i = 0; i < sp1.size(); i++)  {
                    Pos p1 = getPos(sp1[i]);
                    Pos p2 = getPos(sp2[i]);
                    p2[0]+=drawOffset;

                    glLineWidth(1.0);
                    glBegin(GL_LINES);
                    helper::gl::glVertexT(p1);
                    helper::gl::glVertexT(p2);
                    glEnd();
                }
            }
        }
    }
    
    for (size_t seg = 0; seg < bezierSegments.size(); seg++) {
        for (size_t cv = 0; cv < bezierSegments[seg].size(); cv++) {
            Point4 pPrev = bezierSegments[seg][cv].getPoint(0.0);
            Point4 pAct;

            if (vparams->displayFlags().getShowCollisionModels()  &&  cPoints > 0 && d_showCurve.getValue() > 0.0) {
                for (int p = 1; p <= cPoints; p++) {
                    
                    double t = double(p)/double(cPoints);
                    pAct = bezierSegments[seg][cv].getPoint(t);
                    
                    Pos pt1(pPrev[0]+drawOffset, pPrev[1], pPrev[2]);
                    Pos pt2(pAct[0]+drawOffset, pAct[1], pAct[2]);
                    
                    //std::cout << " Points: " << pt1 << " " << pt2 << std::endl;
                    
                    glLineWidth(d_showCurve.getValue());
                    glBegin(GL_LINES);                    
                    glColor4f(rcol[seg],gcol[seg],bcol[seg],1);
                    helper::gl::glVertexT(pt1);
                    helper::gl::glVertexT(pt2);
                    glEnd();
                    
                    pPrev = pAct;
                }
            }
            
            
            if (vparams->displayFlags().getShowInteractionForceFields()  && d_showInputPointSize.getValue() > 0.0) {
                //for (size_t p = 0; p < bezierSegments[seg][cv].curvePoints.size(); p++) {
                for (vector<CurvePoint4>::iterator it = bezierSegments[seg][cv].curvePoints.begin(); it != bezierSegments[seg][cv].curvePoints.end(); it++) {
                    Pos pt = getPos(*it);
                    pt[0] += drawOffset;
                    glPointSize(d_showInputPointSize.getValue());
                    glBegin(GL_POINTS);
                    glColor4f(1,1,0.8,1);
                    helper::gl::glVertexT(pt);
                    glEnd();
                }
                
            }
            
            if (d_showControlPointSize.getValue() > 0.0) {
                //for (size_t p = 0; p < bezierSegments[seg][cv].curvePoints.size(); p++) {
                float fi = 0.5;
                for (vector<Point4>::iterator it = bezierSegments[seg][cv].controlPoints.begin(); it != bezierSegments[seg][cv].controlPoints.end(); it++) {
                    Pos pt = getPos(*it);
                    pt[0] += drawOffset;
                    glPointSize(d_showControlPointSize.getValue());
                    glBegin(GL_POINTS);
                    glColor4f(fi,fi,fi,1);
                    helper::gl::glVertexT(pt);
                    glEnd();
                    fi+=0.1;
                }
                
            }
            
            if (d_showTangentSize.getValue() > 0.0) {
                BezierCubic& bz =  bezierSegments[seg][cv];
                
                Point4 pp1 = bz.curvePoints[0];
                Point4 pp2 = bz.curvePoints[bz.curvePoints.size()-1];
                
                Pos p1 = getPos(pp1);
                Pos p2 = getPos(pp2);
                p1[0] += drawOffset;
                p2[0] += drawOffset;
                
                Point4 vv1 = V4Add(pp1, bz.tangents[0]);
                Point4 vv2 = V4Add(pp2, bz.tangents[1]);
                
                Pos v1 = getPos(vv1);
                Pos v2 = getPos(vv2);
                
                Vec<4,float> drawColor(0.4, 0, 0.4,1);
                vparams->drawTool()->drawArrow(p1,v1,d_showTangentSize.getValue(),drawColor);
                
                Vec<4,float> drawColor2(0.6, 0, 0.6,1);
                vparams->drawTool()->drawArrow(p2,v2,d_showTangentSize.getValue(),drawColor2);
                
                /*glBegin(GL_LINES);
                    glColor4f(1,0,0,1);
                    helper::gl::glVertexT(p1);
                    helper::gl::glVertexT(v1);
                glEnd();*/
                
                /*glLineWidth(d_showTangentSize.getValue());
                glBegin(GL_LINES);
                    glColor4f(1,0,0,1);
                    helper::gl::glVertexT(p2);
                    helper::gl::glVertexT(v2);
                glEnd();*/
            }
            
            if (vparams->displayFlags().getShowBehaviorModels()  &&d_showSampledPointsSize.getValue() > 0.0 ) {
                double pointSize = d_showSampledPointsSize.getValue();
                std::vector<RigidCurvePoint4>& sp = bezierSegments[seg].sampledPoints;
                for (size_t i = 0; i < sp.size(); i++)  {
                    Pos p = getPos(sp[i]);
                    p[0]+=drawOffset;
                    Point4 q = sp[i].quat;
                    
                    Quaternion quat(q[0], q[1], q[2], q[3]);
                    
                    Vec3f vecSize(pointSize, pointSize, pointSize);
                    vparams->drawTool()->drawFrame(p, quat, vecSize);
                }
            }
        }
    }
    glLineWidth(1.0);
    glEnable(GL_LIGHTING);
}



}

}

}

#endif
